from Products.CMFCore.utils import getToolByName
from seantis.plonetools import setuphandlers

indexes = [
    ('source', 'FieldIndex'),
]


def import_indexes(context):
    setuphandlers.import_indexes('seantis.dir.events', indexes, context)


def enable_jquerytools_dateinput_js(context):
    # Ensure that jquerytools.dateinput.js is enabled, it might be disabled
    # with plone.app.jquerytools >= 1.7

    try:
        jstool = getToolByName(context, 'portal_javascripts')
    except AttributeError:
        return

    resource_id = '++resource++plone.app.jquerytools.dateinput.js'
    resource = jstool.getResource(resource_id)
    if resource:
        resource.setEnabled(True)

    jstool.cookResources()



import icalendar
import pytz

from datetime import date, datetime, timedelta
from five import grok
from logging import getLogger
from urllib2 import urlopen
from seantis.dir.events.interfaces import (
    IExternalEventCollector,
    IExternalEventSourceIcal,
    NoImportDataException
)

log = getLogger('seantis.dir.events')


class Missing():

    @property
    def dt(self):
        return None


def as_timezone(value, timezone):
    if isinstance(value, datetime):
        if value.tzinfo:
            return timezone.normalize(value)

        return timezone.localize(value)

    return timezone.localize(datetime(value.year, value.month, value.day))


def as_utc(value):
    return as_timezone(value, pytz.UTC)


class EventsSourceIcal(grok.Adapter):
    grok.context(IExternalEventSourceIcal)
    grok.provides(IExternalEventCollector)

    def fetch(self, ical=None):
        try:
            if ical is None:
                ical = urlopen(self.context.url).read()

            calendar = icalendar.Calendar.from_ical(ical)
        except:
            raise NoImportDataException()

        for event in calendar.walk('vevent'):

            e = {}

            uid = event.get('uid', '').encode('utf-8')
            start = event.get('dtstart', Missing()).dt
            end = event.get('dtend', Missing()).dt
            duration = event.get('duration', Missing()).dt

            e['whole_day'] = False
            if type(start) is date and type(end) is date:
                e['whole_day'] = True

            if type(start) is date:
                start = as_timezone(start, pytz.timezone('Europe/Zurich'))
            if type(end) is date:
                end = as_timezone(end, pytz.timezone('Europe/Zurich'))
                end = end - timedelta(microseconds=1)

            if start and end:
                e['start'] = as_utc(start)
                e['end'] = as_utc(end)
            elif start and duration:
                e['start'] = as_utc(start)
                e['end'] = e['start'] + duration
            else:
                log.error('skipping event {} wtih invalid dates'.format(uid))
                continue

            # Unfortunately, we cannot use the timezone given in the ICS, these
            # might have unusable names such as "W. Europe Standard Time"
            # instead of olson names. We do the same here as with guidle and
            # assume Europe/Zurich!
            e['timezone'] = 'Europe/Zurich'

            e['fetch_id'] = self.context.url
            e['last_update'] = event.get('last-modified', Missing()).dt
            e['source_id'] = uid
            e['title'] = event.get('summary', '').encode('utf-8')
            e['locality'] = event.get('location', '').encode('utf-8')
            des = event.get('description', '').encode('utf-8')
            des = self.context.default_description if not des else des
            e['short_description'] = des
            organizer = event.get('organizer', '').encode('utf-8')
            organizer = 'ical@example.com' if not organizer else organizer
            e['submitter'] = organizer
            e['submitter_email'] = organizer

            e['recurrence'] = ''
            e['long_description'] = ''
            e['cat1'] = set()
            e['cat2'] = set()
            e['street'] = ''
            e['housenumber'] = ''
            e['zipcode'] = ''
            e['town'] = ''
            e['location_url'] = ''
            e['longitude'] = None
            e['latitude'] = None
            e['organizer'] = ''
            e['contact_name'] = ''
            e['contact_email'] = ''
            e['contact_phone'] = ''
            e['prices'] = ''
            e['event_url'] = ''
            e['registration'] = ''
            e['image'] = None
            e['attachment_1'] = None
            e['attachment_2'] = None

            yield e



import os

from datetime import datetime, timedelta
from five import grok
from logging import getLogger
from plone.protect import createToken
from plone.synchronize import synchronized
from seantis.dir.base.interfaces import IDirectoryCatalog
from seantis.dir.events.dates import to_utc
from seantis.dir.events.interfaces import (
    IEventsDirectoryItem, IEventsDirectory, IExternalEvent
)
from seantis.dir.events.recurrence import has_future_occurrences
from seantis.plonetools import unrestricted
from threading import Lock

log = getLogger('seantis.dir.events')


class CleanupScheduler(object):

    _lock = Lock()

    def __init__(self):
        self.next_run = 0

    def is_cleaning_instance(self):
        """ Check if we are the instance which cleans up events.

        This is defined with an environment variable via the buildout file.
        """
        return os.getenv('seantis_events_cleanup', False) == 'true'

    def get_next_run(self, now=None):
        if now is None:
            now = datetime.now()

        # Schedule next run tomorrow at 0:30
        days = 1
        if now.hour < 1 and now.minute < 30:
            days = 0
        next_run = datetime(now.year, now.month, now.day) + timedelta(
            days=days, minutes=30)
        return next_run

    @synchronized(_lock)
    def run(self, directory, dryrun=False, force_run=False, now=None):

        if now is None:
            now = datetime.now()

        if not self.is_cleaning_instance():
            return

        if not self.next_run:
            self.next_run = self.get_next_run(now)

        if (now > self.next_run) or force_run:
            self.next_run = self.get_next_run(now)
            self.cleanup_directory(directory, dryrun=dryrun)

    def remove_stale_previews(self, directory, dryrun=False):

        catalog = IDirectoryCatalog(directory)
        query = catalog.catalog.unrestrictedSearchResults

        log.info('searching for stale previews (> 2 days old)')

        past = to_utc(datetime.utcnow() - timedelta(days=2))
        stale_previews = query(
            path={'query': directory.getPhysicalPath(), 'depth': 2},
            object_provides=IEventsDirectoryItem.__identifier__,
            review_state=('preview'),
            modified={'query': past, 'range': 'max'}
        )
        stale_previews = [p.id for p in stale_previews]

        if stale_previews:
            log.info('deleting stale previews -> %s' % str(stale_previews))
            if not dryrun:
                directory.manage_delObjects(stale_previews)
        else:
            log.info('no stale previews found')

        return stale_previews

    def archive_past_events(self, directory, dryrun=False):

        catalog = IDirectoryCatalog(directory)
        query = catalog.catalog

        log.info('archiving past events (> 2 days old)')

        # events are in the past if they have been over for two days
        # (not one, to ensure that they are really over in all timezones)
        past = to_utc(datetime.utcnow() - timedelta(days=2))
        published_events = query(
            path={'query': directory.getPhysicalPath(), 'depth': 2},
            object_provides=IEventsDirectoryItem.__identifier__,
            review_state=('published', ),
            start={'query': past, 'range': 'max'},
            end={'query': past, 'range': 'max'}
        )

        past_events = []

        for event in published_events:
            event = event.getObject()

            assert event.start < past
            assert event.end < past

            # recurring events may be in the past with one of
            # their occurrences in the future
            if not has_future_occurrences(event, past):
                # published events may be imported events
                if not IExternalEvent.providedBy(event):
                    past_events.append(event)

        ids = [p.id for p in past_events]

        if past_events:
            log.info('archiving past events -> %s' % str(ids))

            if not dryrun:
                for event in past_events:
                    event.archive()
        else:
            log.info('no past events found')

        return ids

    def remove_archived_events(self, directory, dryrun=False):

        catalog = IDirectoryCatalog(directory)
        query = catalog.catalog

        log.info('removing archived events (> 30 days old)')

        past = datetime.utcnow() - timedelta(days=30)
        archived_events = query(
            path={'query': directory.getPhysicalPath(), 'depth': 2},
            object_provides=IEventsDirectoryItem.__identifier__,
            review_state=('archived', ),
            start={'query': past, 'range': 'max'},
            end={'query': past, 'range': 'max'}
        )
        archived_events = [e.id for e in archived_events]

        if archived_events:
            log.info('removing archived events -> %s' % str(archived_events))

            if not dryrun:
                directory.manage_delObjects(archived_events)
        else:
            log.info('no archived events to remove')

        return archived_events

    def remove_past_imported_events(self, directory, dryrun=False):

        catalog = IDirectoryCatalog(directory)
        query = catalog.catalog

        log.info('remove past imported events (> 2 days old)')

        # events are in the past if they have been over for two days
        # (not one, to ensure that they are really over in all timezones)
        past = to_utc(datetime.utcnow() - timedelta(days=2))
        imported_events = query(
            path={'query': directory.getPhysicalPath(), 'depth': 2},
            object_provides=IExternalEvent.__identifier__,
            start={'query': past, 'range': 'max'},
            end={'query': past, 'range': 'max'}
        )

        past_events = []

        for event in imported_events:
            event = event.getObject()

            assert event.start < past
            assert event.end < past

            # recurring events may be in the past with one of
            # their occurrences in the future
            if not has_future_occurrences(event, past):
                past_events.append(event)

        past_events = [p.id for p in past_events]

        if past_events:
            log.info('removing past imported events -> %s' % str(past_events))

            if not dryrun:
                directory.manage_delObjects(past_events)
                pass
        else:
            log.info('no past imported events found')

        return past_events

    def cleanup_directory(self, directory, dryrun=True):

        if dryrun:
            log.info('starting dry run cleanup on %s' %
                     directory.absolute_url())
        else:
            log.info('starting real cleanup on %s' % directory.absolute_url())

        self.remove_stale_previews(directory, dryrun)
        self.archive_past_events(directory, dryrun)
        self.remove_archived_events(directory, dryrun)
        self.remove_past_imported_events(directory, dryrun)

        log.info('finished cleanup on %s' % directory.absolute_url())


cleanup_scheduler = CleanupScheduler()


class CleanupView(grok.View):

    grok.name('cleanup')
    grok.context(IEventsDirectory)
    grok.require('zope2.View')

    def render(self):
        self.request.response.setHeader("Content-type", "text/plain")

        # dryrun must be disabled explicitly using &run=1
        dryrun = not self.request.get('run') == '1'
        force_run = bool(self.request.get('force', False))

        # this maintenance feature may be run unrestricted as it does not
        # leak any information and it's behavior cannot be altered by the
        # user. This allows for easy use via cronjobs.
        with unrestricted.run_as('Manager'):
            self.request.set('_authenticator', createToken())
            cleanup_scheduler.run(self.context, dryrun, force_run)

        return u''



import transaction

from datetime import date, datetime
from seantis.dir.events import dates
from seantis.dir.events.tests import IntegrationTestCase

from seantis.dir.events.catalog import (
    LazyList,
    EventOrderIndex,
    attach_reindex_to_transaction,
    ReindexDataManager,
)

from mock import Mock


class TestCatalog(IntegrationTestCase):

    def test_export_with_categories(self):

        self.login_testuser()

        events = [
            self.create_event(title='1', cat1=['a'], cat2=['1']),
            self.create_event(title='2', cat1=['a', 'b'], cat2=['2']),
            self.create_event(title='3', cat1=['b', 'c'], cat2=['2', '3']),
            self.create_event(title='4', cat1=['c'], cat2=['1', '3'])
        ]
        for event in events:
            event.submit()
            event.publish()
        transaction.commit()

        def _get(cat1=-1, cat2=-1):
            term = {}
            if cat1 != -1:
                term['cat1'] = cat1
            if cat2 != -1:
                term['cat2'] = cat2

            return sorted([s.title for s in self.catalog.export(term=term)])

        self.assertEquals(_get(cat1=None), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat1=[]), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat1=''), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat1='a'), ['1', '2'])

        self.assertEquals(_get(cat1=['a']), ['1', '2'])
        self.assertEquals(_get(cat1=['b']), ['2', '3'])
        self.assertEquals(_get(cat1=['c']), ['3', '4'])
        self.assertEquals(_get(cat1=['a', 'b']), ['1', '2', '3'])
        self.assertEquals(_get(cat1=['a', 'c']), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat1=['b', 'c']), ['2', '3', '4'])
        self.assertEquals(_get(cat1=['a', 'b', 'c']), ['1', '2', '3', '4'])

        self.assertEquals(_get(cat2=None), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat2=[]), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat2=''), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat2='1'), ['1', '4'])

        self.assertEquals(_get(cat2=['1']), ['1', '4'])
        self.assertEquals(_get(cat2=['2']), ['2', '3'])
        self.assertEquals(_get(cat2=['3']), ['3', '4'])
        self.assertEquals(_get(cat2=['1', '2']), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat2=['1', '3']), ['1', '3', '4'])
        self.assertEquals(_get(cat2=['2', '3']), ['2', '3', '4'])
        self.assertEquals(_get(cat2=['1', '2', '3']), ['1', '2', '3', '4'])

        self.assertEquals(_get(cat1=None, cat2=None), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat1=[], cat2=[]), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat1='', cat2=''), ['1', '2', '3', '4'])
        self.assertEquals(_get(cat1='a', cat2='1'), ['1'])

        self.assertEquals(_get(cat1=['a', 'c'], cat2=['1']), ['1', '4'])
        self.assertEquals(_get(cat1=['a'], cat2=['2', '3']), ['2'])
        self.assertEquals(_get(cat1=['b', 'c'], cat2=['3', '4']), ['3', '4'])

        self.directory.manage_delObjects(['1', '2', '3', '4'])
        self.portal.manage_delObjects([self.directory.id])
        transaction.commit()



from five import grok

from seantis.plonetools import async
from seantis.dir.events.interfaces import IResourceViewedEvent


# The primary hook to setup maintenance clockservers is the reservation view
# event. It's invoked way too often, but the invocation is fast and it is
# guaranteed to be run on a plone site with seantis.events installed,
# setup and in use. Other events like zope startup and traversal events are
# not safe enough to use if one has to rely on a site being setup.
@grok.subscribe(IResourceViewedEvent)
def on_resource_viewed(event):
    path = '/'.join(event.context.getPhysicalPath())
    async.register(path + '/fetch?run=1', 15 * 60)
    async.register(path + '/cleanup?run=1', 60 * 60)
